#!/usr/bin/env python3
"""
Created on Wed Nov 20 18:55:05 2019

@author: mikhail-matrosov
"""

import re
from pickle import dumps as pdumps


def hash_obj(*args):
    return '_{:x}'.format(abs(hash(pdumps(args))))


def escape(s):
    return str(s).replace("\\", "\\\\").replace("'", "\\'").replace('"', '\\"')


def as_code(o):
    return f'"{escape(o)}"' if isinstance(o, str) else str(o)


def indent(seq, ind=1):
    return ('    '*ind + s for s in seq)


def store_in_locals(obj, _locals, attr_name=None):
    name = hash_obj(obj)
    _locals[name] = getattr(obj, attr_name) if attr_name else obj
    return name


def cg_type(val):
    if val is None:
        return [f'if o is not None:',
                f'    return None, " type must be NoneType"']
    return [('if not (isinstance(o, float) or isinstance(o, int)):'
             if val == 'number' else
            f'if not isinstance(o, {val}):'),
            f'    return None, " type must be {val}"']


def cg_enum(val, _locals):
    name = store_in_locals(val.copy(), _locals)
    try:
        sname = store_in_locals(set(val), _locals)
        return [
             'try:',
            f'    if o not in {sname}:',
            f'        return None, f" must be one of {{{name}}}"',
             'except TypeError:',  # If o is not Hashable
            f'    return None, f" must be one of {{{name}}}"']
    except TypeError:  # Not hashable
        return [f'if o not in {name}:',
                f'    return None, f" must be one of {{{name}}}"']


def cg_min(val):
    return [f'if o < {as_code(val)}:',
            f"    return None, ' must be at least {as_code(val)}'"]


def cg_max(val):
    return [f'if o > {as_code(val)}:',
            f"    return None, ' must be at most {as_code(val)}'"]


def cg_min_len(val):
    return [f'if len(o) < {val}:',
            f'    return None, " length must be at least {val}"']


def cg_max_len(val):
    return [f'if len(o) > {val}:',
            f'    return None, " length must be at most {val}"']


def cg_regex(val, _locals):
    name = store_in_locals(re.compile(val), _locals, 'fullmatch')
    return [f'if not {name}(o):',
            f'    return None, " must match regex: {{}}".format({name}.__self__.pattern)']


def cg_coerce_int():
    return ['try: o = int(o)',
            'except:',
            '    try: o = int(float(o))',
            '    except:',
            '        return None, " is not coercible to int"']


def cg_coerce_float():
    return ['try: o = float(o)',
            'except:',
            '    return None, " is not coercible to float"']


def cg_coerce_number():
    return ['if not isinstance(o, float):',
            '    try: o = int(o)',
            '    except:',
            '        try: o = float(o)',
            '        except:',
            '            return None, " is not coercible to number"']


def cg_coerce_str():
    yield 'o = str(o)'


def cg_map(val, _locals):
    name = store_in_locals(val.copy(), _locals, 'get')
    default = f'{name}(None)' if None in val else 'o'
    return [f'try: o = {name}(o, {default})',
             'except TypeError: pass']  # In case not Hashable


def cg_any_of(names):
    for i, name in enumerate(names):
        yield 4*i*' ' + f'r{i}, err{i} = {name}(o, args)'
        yield 4*i*' ' + f'if err{i}:'

    N = len(names)
    errs = ', '.join(f'_indent_str(err{i})' for i in range(N))
    fmt = ''.join(f'\\n{i+1}: ^{{}}' for i in range(N))
    yield 4*N*' ' + f'return None, " must satisfy any of {N} rules:{fmt}".format({errs})'
    yield from (4*i*' ' + f'else: o = r{i}' for i in reversed(range(N)))


def cg_one_of(names):
    N = len(names)
    errs = '"^"+_indent_str(e) if e else "ok" for x, e in rs'
    fmt = ''.join(f'\\n{i+1}: {{}}' for i in range(N))
    calls = ', '.join(f'{name}(o, args)' for name in names)
    return [
    f'rs = [{calls}]',
     'valid = [x for x, e in rs if not e]',
     'if len(valid) == 1:',
     '    o = valid[0]',
     'else:',
    f'    fmt = ({errs})',
    f'    return None, " must satisfy exactly one of {N} rules:{fmt}".format(*fmt)']


def cg_coerce(name, input_code='o'):
    try:
        return globals()['cg_coerce_'+name]()
    except KeyError:
        return [
         'try:',
        f'    o = self.coerce_{name}({input_code}, args)',
         'except:',
        f'    if not hasattr(self, "coerce_{name}"):',
        f'        raise AttributeError("Validator has no attribute coerce_{name}")',
        f'    return None, " is not coercible to {name}"']


def cg_nullable(if_null, post_coerce):
    if isinstance(if_null, str) and re.match('{.+}', if_null):
        if_null = f'args[{as_code(if_null[1:-1])}]'
    else:
        if_null = as_code(if_null)
    yield 'if o is None:'
    if post_coerce:
        yield from cg_coerce(post_coerce, if_null)
    else:
        yield f'    return {if_null}, None'


def cg_rules(self, rules, options):
    _locals = self.__dict__
    rules = rules or {}
    rget = rules.get
    globs = globals()
    resolve = self.resolve_rules
    rhash, rules = resolve(rules, options)

    # Resolve rules from 'rules' rule
    v = rget('rules')
    if v:
        old_rules, rules = rules, resolve(v)[1].copy()
        rules.update(old_rules)
        rget = rules.get

    if rget('examples'):
        self._positive_examples[rhash] = (rules, rget('examples'))

    if rget('negative_examples'):
        self._negative_examples[rhash] = (rules, rget('negative_examples'))

    if rget('nullable') or rget("if_null"):
        yield from cg_nullable(rget("if_null"), rget('post_coerce'))

    if 'type' in rules:
        yield from cg_type(rules['type'])

    v = rget('coerce')
    if v:
        yield from cg_coerce(v)

    v = rget('regex')
    if v:
        yield from cg_regex(v, _locals)

    purge_unkown = rget('purge_unkown', options.purge_unkown)
    has_schema = any(map(rget, 'items values keys pattern_items'.split()))

    rtype = rget('type', rget('coerce'))
    if rtype in {'dict', 'list'} or has_schema:
        if rtype == 'dict' and purge_unkown:
            yield 'o, orig = {}, o'
        else:
            yield 'o, orig = o.copy(), o'

    inner_options = self.options.replace(**rules)

    if has_schema:
        if rtype == 'list':
            yield from cg_list_items(self, **rules, options=inner_options)
        else:
            yield from cg_dict_items(self, **rules, options=inner_options)

    v = rget('map')
    if v:
        yield from cg_map(v, _locals)

    v = rget('enum')
    if v:
        yield from cg_enum(v, _locals)

    for r in ['min', 'max', 'min_len', 'max_len']:
        v = rget(r)
        if v is not None:
            yield from globs['cg_' + r](v)

    for k in ['any_of', 'one_of']:
        v = rget(k)
        if v:
            names = [resolve(r, inner_options)[0] for r in v]
            yield from globs['cg_' + k](names)

    v = rget('post_coerce')
    if v:
        yield from cg_coerce(v)


def cg_default(k, k_to, rules, require_all):
    if 'default' in rules:
        v = rules['default']
        if isinstance(v, str) and re.fullmatch('{.+}', v):
            default = f'args[{as_code(v[1:-1])}]'
        else:
            default = as_code(v)

        yield  'else:'
        yield f'    o[{k_to}] = {default}'
    elif rules.get('required', require_all):
        yield  'else:'
        yield f'    return None, ".{escape(k)} is required"'


def cg_key_value(k, fname, rules, require_all):
    k_from = as_code(k)
    k_to = as_code(rules['rename']) if rules and 'rename' in rules else k_from
    val_source = (f'o.pop({k_from}, orig[{k_from}])'
                  if 'rename' in rules else f'orig[{k_from}]')
    yield from [
        f'if {k_from} in orig:',
        f'    o[{k_to}], err = {fname}({val_source}, args)',
         '    if err:',
        f'        return None, ".{escape(k)}" + err']

    yield from cg_default(k, k_to, rules, require_all)


def cg_dict_item(self, key, rules, require_all, store_known_keys):
    fname, rules = self.resolve_rules(rules)
    k_from = as_code(key)
    k_to = as_code(rules['rename']) if rules and 'rename' in rules else k_from

    if rules is NotImplemented:
        yield from cg_key_value(key, fname, {}, require_all)
    elif rules:
        synonyms = rules.get('synonyms')
        if synonyms:
            store_known_keys.update(synonyms)
#            val_source = ('o.pop(k, orig[k])'
#                          if 'rename' in rules else 'orig[k]')
            val_source = 'o.pop(k, orig[k])'
            yield from [
                f'for k in {[key] + synonyms}:',
                 '    if k in orig:',
                f'        o[{k_to}], err = {fname}({val_source}, args)',
                 '        if err:',
                 '            return None, f".{k}{err}"',
                 '        break']
            yield from cg_default(key, k_to, rules, require_all)
        else:
            yield from cg_key_value(key, fname, rules, require_all)
    else:
        yield f'if {k_from} in orig:'
        yield f'    o[{k_to}] = orig[{k_from}]'
        yield from cg_default(key, k_to, {}, require_all)


def cg_rule_key_value(self, keys, values, err_key_fmt='.{}'):
    if keys is None:
        k_to, ind = 'k', ''
    else:
        k_fname, k_rules = self.resolve_rules(keys)
        if k_rules == {}:
            k_to, ind = 'k', ''
        else:
            yield f'tk, k_err = {k_fname}(k, args)'
            yield  'if not k_err:'
            k_to, ind = 'tk', '    '

    if values:
        v_fname, v_rules = self.resolve_rules(values)
        if v_rules == {}:
            yield ind + f'o[{k_to}] = orig[k]'
        else:
            yield from [
            ind + f'o[{k_to}], err = {v_fname}(orig[k], args)',
            ind +  'if err:',
            ind + f'    return None, "{err_key_fmt}{{}}".format(k, err)',
            ind +  'continue']
    else:
        yield ind + f'o[{k_to}] = orig[k]'


def cg_pattern_item(self, pattern_items):
    _locals = self.__dict__
    resolve = self.resolve_rules

    yield 'if isinstance(k, str):'

    for ptrn in sorted(pattern_items, key=lambda p: (-len(p), p)):
        fname, rules = resolve(pattern_items[ptrn])

        # TODO: check if rules is NotImplemented
        name = store_in_locals(re.compile(ptrn), _locals, 'fullmatch')
        tk = as_code(rules['rename']) if rules and 'rename' in rules else 'k'

        yield from [
            f'    if {name}(k):',
            f'        o[{tk}], err = {fname}(orig[k], args)',
             '        if err:',
             '            return None, f".{k}{err}"',
             '        continue']
        # TODO: required and default


def cg_dict_items(self, items=None, pattern_items=None, keys=None, values=None,
                  options=None, **_):
    known_keys = set(items) if items else set()
    _locals = self.__dict__

    if items:
        for key, rules in items.items():
            yield from cg_dict_item(self, key, rules, options.require_all, known_keys)

    pattern_items = pattern_items or {}
    if pattern_items or keys is not None or values is not None:
        if known_keys:
            kk_name = store_in_locals(known_keys, _locals)
            yield f'for k in set(orig) - {kk_name}:'
        else:
            yield 'for k in orig:'

        if pattern_items:
            yield from indent(cg_pattern_item(self, pattern_items))

        if keys is not None or values is not None:
            yield from indent(cg_rule_key_value(self, keys, values))

        if not options.allow_unkown:
            yield '    return None, f" must not contain key {k}"'

    elif not options.allow_unkown:
        kk_name = store_in_locals(known_keys, _locals)
        yield from [
        f'forbidden_keys = set(orig) - {kk_name}',
         'if forbidden_keys:',
         '    return None, f" must not contain keys {forbidden_keys}"']


def cg_list_items(self, keys=None, values=None, **_):
    yield 'for k in range(len(orig)):'

    if values is not None:
        yield from indent(cg_rule_key_value(
                self, keys, values, err_key_fmt='[{}]'))




#!/usr/bin/env python3
"""
Created on Wed Nov 20 19:35:46 2019

@author: mikhail-matrosov
"""

from functools import wraps
from pycoercer.code_generator import hash_obj, indent, cg_rules


def log(*a, **kw):
    # print(*a, **kw)
    pass


def loopbreaker(f):
    locks = set()

    @wraps(f)
    def wrapper(doc, args):
        k = id(doc)
        if k in locks:
            return doc, None

        locks.add(k)
        try:
            return f(doc, args)
        finally:  # Do after return
            locks.remove(k)

    wrapper.locks = locks
    return wrapper


def indent_str(s):
    return '\n  '.join(s.splitlines()) if '\n' in s else s


def schema_not_found(name):
    def f(o, args):
        raise NameError(f"Schema {name} was not defined")
    f.__name__ = name
    f.n_lines = 0
    return f


def _compile(statements, name, _locals, schema, break_loops=False):
    f_cache = _locals.setdefault('_f_cache', {})

    h = hash_obj(statements)
    if h in f_cache:
        f = _locals[name] = f_cache[h]
        log(f'# {name} = {f.__name__}\n')
        return f

    src = '\n'.join((f'def {name}(o, args):',
                     *indent(statements),
                     '    return o, None'))

    log(src + '\n')
    exec(src, _locals)

    func = _locals[name]
    if break_loops:
        func = loopbreaker(func)

    func.src = src
    func.schema = schema

    f_cache[h] = _locals[name] = func
    return func


def registry_wrapper(f):
    @wraps(f)
    def wrapper(obj, args=None):
        doc, err = f(obj, args)
        if err:
            return doc, 'Input' + err
        return doc, None

    return wrapper


class BasicValidator:
    def __init__(self):
        self._todo = []
        self._schemas = {}
        self._positive_examples = {}
        self._negative_examples = {}

        # Used by generated code
        self.self = self
        self._indent_str = indent_str

    def generate_function(self, rules, options, name):
        _locals = self.__dict__
        fname, rules = self.resolve_rules(rules, options)

        log(f'# {name}\n# {rules}')
        statements = list(cg_rules(self, rules, options))
        func = _compile(statements, fname, _locals, rules, break_loops=True)

        _locals[hash_obj(name)] = func  # Link previously loaded refs to f

        compiled_set = set()

        for task, opts in self._todo:
            fname = hash_obj(task, opts.__dict__)
            if fname not in compiled_set:  # Avoid recursion
                compiled_set.add(fname)
                log(f'# {task}')
                statements = list(cg_rules(self, task, opts))
                _compile(statements, fname, _locals, task)

        self._todo.clear()

        return registry_wrapper(func)

    def _test_examples(self):
        _locals = self.__dict__
        for fname, (rules, examples) in self._positive_examples.items():
            func = _locals[fname]
            for example in examples:
                if func(example, None)[1]:
                    raise ValueError('Failed validating example '
                                     f'{example} agains schema {rules}')

        for fname, (rules, examples) in self._negative_examples.items():
            func = _locals[fname]
            for example in examples:
                if not func(example, None)[1]:
                    raise ValueError('Failed validating negative example '
                                     f'{example} agains schema {rules}')

    def resolve_rules(self, rules, options=None):
        '''
        Idempotent
        return hash, {} - no schema to check
        return hash, NotImplemented - schema might come later
        '''

        if not options:
            options = self.options

        h = hash_obj(rules) if isinstance(rules, str) else hash_obj(rules, options.__dict__)

        if rules is NotImplemented:
            return h, NotImplemented

        if isinstance(rules, str):
            try:
                return h, (self._schemas[rules] or {})
            except KeyError:
                # Create a stub
                self.__dict__[h] = schema_not_found(rules)
                return h, NotImplemented
        elif rules:  # Avoid empty rulesets
            self._todo.append((rules, options))
            return h, rules
        return h, {}



#!/usr/bin/env python3
"""
Created on Fri Sep 20 12:37:07 2019

@author: mikhail-matrosov
"""

from pycoercer.basic_validator import BasicValidator


class Options():
    def __init__(self,
                 allow_unkown=True,
                 purge_unkown=False,
                 require_all=False,
                 break_loops=True,
                 load_as_jsonschema=False,
                 validate_schemas=True,
                 **_):
        self.allow_unkown = allow_unkown
        self.purge_unkown = purge_unkown
        self.require_all = require_all
        self.break_loops = break_loops  # Makes Phimera ~10-15% slower
        self.load_as_jsonschema = load_as_jsonschema
        self.validate_schemas = validate_schemas

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def replace(self, **kwargs):
        '''Returns a new instance'''
        data = self.__dict__.copy()
        data.update(kwargs)
        return Options(**data)


class Validator(BasicValidator):
    def __init__(self, schemas: dict = None, options=None, **kwargs):
        super().__init__()
        self.registry = {}
        self.options = (options or Options()).replace(**kwargs)

        if schemas:
            self.update(schemas)

    def __getitem__(self, k):
        return self.registry[k]

    def __setitem__(self, key, schema: dict):
        self.update({key: schema})

    def update(self, schemas: dict, options=None, **kwargs):
        options = (options or self.options).replace(**kwargs)
        self.options, options_backup = options, self.options

        if options.load_as_jsonschema:
            schemas = {k: {'type': 'dict', 'schema': v}
                       for k, v in schemas.items()}

        # Validate input schemas
        if options.validate_schemas:
            schemas, err = pycoercer_schema_validator(schemas)
            if err:
                raise ValueError(err)

        self._schemas.update(schemas)

        # Code generation
        self.registry.update({
            name: self.generate_function(schema, options, name)
            for name, schema in schemas.items()
        })

        # Validate examples
        try:
            if options.validate_schemas:
                self._test_examples()
        finally:  # even if exception
            self._positive_examples.clear()
            self._negative_examples.clear()

        self.options = options_backup


pycoercer_schema = {
    'str': {'type': 'str'},
    'int': {'type': 'int'},
    'bool': {'type': 'bool'},

    'rules': {
        'type': 'dict',
        'items': {
            'title': None,
            'description': None,

            'examples': {'type': 'list'},
            'negative_examples': {'type': 'list'},

            'allow_unkown': 'bool',
            'purge_unkown': 'bool',

            'rename': 'str',
            'synonyms': {'type': 'list'},
            'required': 'bool',
            'require_all': 'bool',

            'nullable': 'bool',
            'if_null': {},
            'default': {},

            'type': {
                'nullable': True,
                'type': 'str',
                'map': {
                    'object': 'dict',
                    'array': 'list',
                    'string': 'str',
                    'integer': 'int',
                    'boolean': 'bool',
                    'None': None,
                    'null': None
                },
                'enum': ['dict', 'list', 'str', 'int', 'float', 'number',
                         'bool']
            },
            'coerce': 'str',
            'map': {
                'type': 'dict',
                'allow_unkown': True
            },
            'enum': {
                'type': 'list',
                'synonyms': ['allowed']
            },
            'regex': {
                'type': 'str',
                'synonyms': ['pattern']
            },
            'items': {
                'type': 'dict',  # TODO: list notation for lists
                'values': 'obj',
                'synonyms': ['schema', 'properties']
            },
            'rules': {'type': 'str'},
            'keys': {
                'rules': 'obj',
                'synonyms': ['keysrules']
            },
            'values': {
                'rules': 'obj',
                'synonyms': ['valuesrules']
            },
            'min': {},
            'max': {},
            'min_len': {
                'type': 'int',
                'synonyms': ['minLength', 'minlength']
            },
            'max_len': {
                'type': 'int',
                'synonyms': ['maxLength', 'maxlength']
            },
            'one_of': {
                'type': 'list',
                'values': 'obj',
                'synonyms': ['oneOf', 'oneof']
            },
            'any_of': {
                'type': 'list',
                'values': 'obj',
                'synonyms': ['anyOf', 'anyof']
            },
            'post_coerce': 'str'

            # todo: if_invalid
        }
    },

    'obj': {
        'any_of': [
            {'type': None},
            'str',
            'rules'
        ]
    },

    'obj_dict': {
        'type': 'dict',
        'values': 'obj'
    }
}

_pcsv = Validator(
    pycoercer_schema,
    allow_unkown=False,
    purge_unkown=False,
    require_all=False,
    break_loops=True,
    load_as_jsonschema=False,
    validate_schemas=False)

pycoercer_schema_validator = _pcsv['obj_dict']



#!/usr/bin/env python3
"""
Created on Mon Nov 25 17:31:20 2019

@author: mikhail-matrosov
"""

from distutils.core import setup


try:
    LONG_DESCRIPTION = open('README.rst').read()
except:
    LONG_DESCRIPTION = ''

setup(
  name = 'pycoercer',
  packages = ['pycoercer'],
  version = '0.1',
  license = 'MIT',
  description = 'Fast Python JSON schema validation and normalization',
  long_description = LONG_DESCRIPTION,
  author = 'Mikhail Matrosov',
  author_email = 'mm@tardis3d.ru',
  url = 'https://github.com/mikhail-matrosov/pycoercer',
  download_url = 'https://github.com/mikhail-matrosov/pycoercer/releases/download/0.1/pycoercer-0.1.tar.gz',
  keywords = ['validation', 'schema', 'json', 'normalization', 'coercion'],
  install_requires = [],
  platforms=["any"],
  tests_require=["pytest"],

  classifiers = [
    'Development Status :: 3 - Alpha',  # 3 - Alpha, 4 - Beta or 5 - Production/Stable
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ]
)
